from decimal import Decimal

from django.db.models import QuerySet, Sum
from wbcore import viewsets
from wbcore.utils.strings import format_number

from wbaccounting.models import BookingEntry
from wbaccounting.models.booking_entry import BookingEntryType
from wbaccounting.serializers import (
    BookingEntryModelSerializer,
    BookingEntryRepresentationSerializer,
    BookingEntryTypeModelSerializer,
    BookingEntryTypeRepresentationSerializer,
)
from wbaccounting.viewsets.buttons import BookingEntryButtonConfig
from wbaccounting.viewsets.display import BookingEntryDisplayConfig
from wbaccounting.viewsets.titles import BookingEntryTitleConfig


class BookingEntryTypeRepresentationViewSet(viewsets.RepresentationViewSet):
    search_fields = ["key", "title"]
    ordering_fields = ["title"]
    ordering = ["title"]
    serializer_class = BookingEntryTypeRepresentationSerializer
    queryset = BookingEntryType.objects.all()


class BookingEntryRepresentationViewSet(viewsets.RepresentationViewSet):
    search_fields = ["counterparty__computed_str", "title", "currency__key"]
    ordering_fields = ["title"]
    ordering = ["title"]
    serializer_class = BookingEntryRepresentationSerializer
    queryset = BookingEntry.objects.all()

    def get_queryset(self):
        return BookingEntry.objects.filter_for_user(self.request.user)


class BookingEntryTypeModelViewSet(viewsets.ModelViewSet):
    search_fields = ["key", "title"]
    ordering_fields = ["title"]
    ordering = ["title"]
    serializer_class = BookingEntryTypeModelSerializer
    queryset = BookingEntryType.objects.all()


class BookingEntryModelViewSet(viewsets.ModelViewSet):
    filterset_fields = {
        "counterparty": ["exact"],
        "booking_date": ["gte", "exact", "lte"],
        "reference_date": ["gte", "exact", "lte"],
    }
    search_fields = ["counterparty__computed_str", "title", "currency__key"]
    ordering_fields = (
        "counterparty__computed_str",
        "booking_date",
        "gross_value",
        "net_value",
        "reference_date",
    )
    ordering = ["-booking_date"]

    serializer_class = BookingEntryModelSerializer
    queryset = BookingEntry.objects.all()

    button_config_class = BookingEntryButtonConfig
    display_config_class = BookingEntryDisplayConfig
    title_config_class = BookingEntryTitleConfig

    def get_aggregates(self, queryset, paginated_queryset):
        if not queryset.exists():
            return {}

        return {
            "net_value": {"Σ": format_number(queryset.aggregate(s=Sum("net_value"))["s"] or Decimal("0"), decimal=8)},
            "gross_value": {
                "Σ": format_number(queryset.aggregate(s=Sum("gross_value"))["s"] or Decimal("0"), decimal=8)
            },
            "invoice_net_value": {
                "Σ": format_number(queryset.aggregate(s=Sum("invoice_net_value"))["s"] or Decimal("0"), decimal=8)
            },
            "invoice_gross_value": {
                "Σ": format_number(queryset.aggregate(s=Sum("invoice_gross_value"))["s"] or Decimal("0"), decimal=8)
            },
        }

    def get_queryset(self) -> QuerySet[BookingEntry]:
        booking_entries = BookingEntry.objects.filter_for_user(self.request.user)

        if eai_id := self.kwargs.get("entry_accounting_information_id", None):
            booking_entries = booking_entries.filter(counterparty__entry_accounting_information__id=eai_id)

        if invoice_id := self.kwargs.get("invoice_id", None):
            booking_entries = booking_entries.filter(invoice_id=invoice_id)

        return booking_entries.select_related(
            "counterparty",
            "invoice",
            "product",
            "type",
            "account",
            "currency",
            "invoice__invoice_currency",
        )



# Generated by Django 5.2.9 on 2026-01-21 12:55
from datetime import date

import django.contrib.postgres.constraints
import django.contrib.postgres.fields.ranges
import django.db.models.deletion
from django.db import migrations, models
from psycopg.types.range import DateRange

def migrate_vat(apps, schema_editor):
    EntryAccountingInformation = apps.get_model('wbaccounting', 'EntryAccountingInformation')
    Vat = apps.get_model('wbaccounting', 'Vat')
    VatRate = apps.get_model('wbaccounting', 'VatRate')
    objs = []
    default_vat = Vat.objects.create(name=str(0.08), primary=True)
    VatRate.objects.create(rate=0.08, vat=default_vat, timespan=DateRange(date(2010, 1, 1), None))
    for entry in EntryAccountingInformation.objects.all():
        if entry.old_vat:
            try:
                rate = VatRate.objects.get(rate=entry.old_vat)
                entry.vat = rate.vat
            except VatRate.DoesNotExist:
                vat = Vat.objects.create(name=str(entry.old_vat))
                rate = VatRate.objects.create(rate=entry.old_vat, vat=vat, timespan=DateRange(date(2010,1,1), None))
                entry.vat = rate.vat
        else:
            entry.vat = default_vat
        objs.append(entry)
    EntryAccountingInformation.objects.bulk_update(objs, ["vat"])

class Migration(migrations.Migration):

    dependencies = [
        ('wbaccounting', '0012_entryaccountinginformation_external_invoice_users'),
    ]

    operations = [
        migrations.CreateModel(
            name='Vat',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ("primary", models.BooleanField(default=False, verbose_name='Primary'))
            ],
        ),
        migrations.RenameField(model_name='entryaccountinginformation', old_name="vat", new_name="old_vat",),
        migrations.AddField(
            model_name='entryaccountinginformation',
            name='vat',
            field=models.ForeignKey(default=None, blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='vat_recipients', to='wbaccounting.vat', verbose_name='VAT'),
            preserve_default=False,
        ),
        migrations.CreateModel(
            name='VatRate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('timespan', django.contrib.postgres.fields.ranges.DateRangeField(verbose_name='Timespan')),
                ('rate', models.DecimalField(decimal_places=4, max_digits=5)),
                ('vat', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='rates', to='wbaccounting.vat')),
            ],
            options={
                'verbose_name': 'Vat Rate',
                'verbose_name_plural': 'Vat Rates',
                'constraints': [django.contrib.postgres.constraints.ExclusionConstraint(expressions=[('timespan', '&&'), ('vat', '=')], name='exclude_overlapping_vat_rates')],
            },
        ),
        migrations.RunPython(migrate_vat),
        migrations.RemoveField(model_name='entryaccountinginformation', name="old_vat"),
        migrations.AlterField(
            model_name='entryaccountinginformation',
            name='vat',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.PROTECT,
                                    related_name='vat_recipients', to='wbaccounting.vat', verbose_name='VAT'),
            preserve_default=False,
        ),
        migrations.AlterModelOptions(
            name='vat',
            options={'verbose_name': 'Vat Group', 'verbose_name_plural': 'Vat Groups'},
        ),
        migrations.AddConstraint(
            model_name='vat',
            constraint=models.UniqueConstraint(fields=('primary',), name='unique_primary_vat_group'),
        ),
    ]



from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class BookingEntryDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> dp.ListDisplay:
        _counterparty_fields = [dp.Field(key="counterparty", label="Counterparty")]
        _common_fields = [
            dp.Field(
                key=None,
                label="Product",
                open_by_default=False,
                children=[
                    dp.Field(key="product", label="Product", width=325),
                    dp.Field(key="type", label="Type", width=250),
                    dp.Field(key="account", label="Account", width=250, show="open"),
                ],
            ),
            dp.Field(
                key=None,
                label="Date",
                open_by_default=False,
                children=[
                    dp.Field(key="period", label="Period", width=250),
                    dp.Field(key="booking_date", label="Book. Date", width=100),
                    dp.Field(key="reference_date", label="Ref. Date", width=100, show="open"),
                    dp.Field(key="payment_date", label="Payment. Date", width=100, show="open"),
                ],
            ),
            dp.Field(
                key=None,
                label="Value",
                open_by_default=False,
                children=[
                    dp.Field(key="gross_value", label="Gross", width=100),
                    dp.Field(key="net_value", label="Net", width=100),
                    dp.Field(key="vat", label="VAT", width=100, show="open"),
                    dp.Field(key="quantity", label="Quantity", width=100, show="open"),
                ],
            ),
            dp.Field(
                key=None,
                label="Invoice Value",
                open_by_default=False,
                children=[
                    dp.Field(key="invoice_fx_rate", label="FX", width=100),
                    dp.Field(key="invoice_gross_value", label="Gross", width=100, show="open"),
                    dp.Field(key="invoice_net_value", label="Net", width=100, show="open"),
                ],
            ),
        ]
        _invoice_fields = [dp.Field(key="invoice", label="Invoice", width=325)]

        if "invoice_id" in self.view.kwargs:
            return dp.ListDisplay(fields=_common_fields)

        if "entry_accounting_information_id" in self.view.kwargs:
            return dp.ListDisplay(fields=[*_common_fields, *_invoice_fields])

        return dp.ListDisplay(fields=[*_counterparty_fields, *_common_fields, *_invoice_fields])

    def get_instance_display(self) -> dp.Display:
        return dp.Display(
            pages=[
                dp.Page(
                    layouts={
                        dp.default(): dp.Layout(
                            grid_template_areas=[
                                ["title", "title", "title", "title"],
                                ["product", "product", "type", "type"],
                                ["counterparty", "counterparty", "period", "period"],
                                ["booking_date", "reference_date", "due_date", "payment_date"],
                                ["currency", "vat", "gross_value", "net_value"],
                                ["invoice_currency", "invoice_fx_rate", "invoice_gross_value", "invoice_net_value"],
                                ["invoice", "invoice", "account", "account"],
                            ],
                            grid_template_columns=[
                                "200px",
                                "200px",
                                "200px",
                                "200px",
                            ],
                        )
                    }
                )
            ]
        )



# Generated by Django 5.2.9 on 2026-02-12 13:25
from decimal import Decimal

import django.contrib.postgres.constraints
import django.contrib.postgres.fields.ranges
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('currency', '0002_alter_currencyfxrates_unique_together_and_more'),
        ('directory', '0015_alter_emailcontact_address_and_more'),
        ('wbaccounting', '0014_bookingentrytype_remove_bookingentry_title_and_more'),
        ('wbcrm', '0020_alter_product_unique_together_product_unique_product'),
    ]

    operations = [
        migrations.AlterField(
            model_name='bookingentry',
            name='backlinks',
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name='bookingentry',
            name='parameters',
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name='bookingentry',
            name='period',
            field=django.contrib.postgres.fields.ranges.DateRangeField(default=None, verbose_name='Period'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='bookingentry',
            name='key',
            field=models.CharField(default=None, max_length=255, verbose_name='Key'),
            preserve_default=False,
        ),
        migrations.RemoveField(
            model_name='invoice',
            name='gross_value',
        ),
        migrations.RemoveField(
            model_name='invoice',
            name='net_value',
        ),
        migrations.RemoveField(
            model_name='bookingentry',
            name='invoice_gross_value',
        ),
        migrations.RemoveField(
            model_name='bookingentry',
            name='invoice_net_value',
        ),
        migrations.AddField(
            model_name='bookingentry',
            name='invoice_gross_value',
            field=models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(
                models.F('invoice_fx_rate'), '*', models.F('gross_value')),
                                        output_field=models.DecimalField(blank=True, decimal_places=4, max_digits=16,
                                                                         null=True)),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='bookingentry',
            name='invoice_net_value',
            field=models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(models.F('invoice_fx_rate'), '*', django.db.models.expressions.CombinedExpression(models.F('gross_value'), '/', django.db.models.expressions.CombinedExpression(models.Value(Decimal('1')), '+', models.F('vat')))), output_field=models.DecimalField(blank=True, decimal_places=4, max_digits=16, null=True)),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='bookingentry',
            name='type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE,
                                    related_name='booking_entries', to='wbaccounting.bookingentrytype',
                                    verbose_name='Type'),
        ),
        migrations.AlterField(
            model_name='bookingentry',
            name='gross_value',
            field=models.DecimalField(decimal_places=4, default=Decimal('0'), max_digits=16,
                                      verbose_name='Gross Value'),
        ),
        migrations.RemoveField(
            model_name='bookingentry',
            name='net_value',
        ),
        migrations.AddField(
            model_name='bookingentry',
            name='net_value',
            field=models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(models.F('gross_value'), '/', django.db.models.expressions.CombinedExpression(models.Value(Decimal('1')), '+', models.F('vat'))), output_field=models.DecimalField(decimal_places=4, max_digits=16)),
        ),
        migrations.AddIndex(
            model_name='bookingentry',
            index=models.Index(fields=['counterparty', 'type', 'product', 'account', 'period'],
                               name='wbaccountin_counter_565eaa_idx'),
        ),
        migrations.AddIndex(
            model_name='bookingentry',
            index=models.Index(fields=['counterparty', 'type', 'product', 'account', 'period', 'invoice'],
                               name='wbaccountin_counter_70bf5b_idx'),
        ),
        migrations.AddConstraint(
            model_name='bookingentry',
            constraint=django.contrib.postgres.constraints.ExclusionConstraint(
                condition=models.Q(('invoice__isnull', True), ('account__isnull', False)),
                expressions=[('period', '&&'), ('counterparty', '='), ('key', '='), ('account', '=')],
                name='exclude_overlapping_booking_entry'),
        ),
        migrations.AddConstraint(
            model_name='bookingentry',
            constraint=django.contrib.postgres.constraints.ExclusionConstraint(
                condition=models.Q(('invoice__isnull', True), ('account__isnull', True)),
                expressions=[('period', '&&'), ('counterparty', '='), ('key', '=')],
                name='exclude_overlapping_booking_entry_no_account'),
        ),
        migrations.AddConstraint(
            model_name='bookingentry',
            constraint=django.contrib.postgres.constraints.ExclusionConstraint(
                condition=models.Q(('invoice__isnull', False), ('account__isnull', False)),
                expressions=[('period', '&&'), ('counterparty', '='), ('key', '='), ('account', '='), ('invoice', '=')],
                name='exclude_overlapping_invoiced_booking_entry'),
        ),
        migrations.AddConstraint(
            model_name='bookingentry',
            constraint=django.contrib.postgres.constraints.ExclusionConstraint(
                condition=models.Q(('invoice__isnull', False), ('account__isnull', True)),
                expressions=[('period', '&&'), ('counterparty', '='), ('key', '='), ('invoice', '=')],
                name='exclude_overlapping_invoiced_booking_entry_no_account'),
        ),

    ]



from django.template import Context, Template
from reportlab.lib import colors, utils
from reportlab.lib.enums import TA_CENTER, TA_RIGHT
from reportlab.lib.pagesizes import A4
from reportlab.lib.styles import ParagraphStyle, getSampleStyleSheet
from reportlab.lib.units import cm
from reportlab.platypus import Image, Table, TableStyle
from wbcore.utils.reportlab import FormattedParagraph as Paragraph
from wbcrm.models import Account

from wbaccounting.dynamic_preferences_registry import format_invoice_number


def get_styles():
    styles = getSampleStyleSheet()
    styles.add(ParagraphStyle(name="Headline", fontName="Helvetica", fontSize=24, leading=34))
    styles.add(ParagraphStyle(name="Default", fontName="Helvetica", fontSize=10, leading=12))
    styles.add(ParagraphStyle(name="Default-Table-Header", fontName="Helvetica", fontSize=14, leading=16))
    styles.add(
        ParagraphStyle(
            name="Default-Table-Header-Right", fontName="Helvetica", fontSize=14, leading=16, alignment=TA_RIGHT
        )
    )
    styles.add(
        ParagraphStyle(
            name="Default-Right", fontName="Helvetica", fontSize=9, leading=12, alignment=TA_RIGHT, splitLongWords=0
        )
    )
    styles.add(
        ParagraphStyle(
            name="Default-Right-Linebreak",
            fontName="Helvetica",
            fontSize=10,
            leading=12,
            alignment=TA_RIGHT,
            spaceAfter=12,
        )
    )
    styles.add(
        ParagraphStyle(name="Default-Center", fontName="Helvetica", fontSize=9, leading=12, alignment=TA_CENTER)
    )
    styles.add(ParagraphStyle(name="Default-Linebreak", fontName="Helvetica", fontSize=10, leading=12, spaceAfter=12))
    styles.add(
        ParagraphStyle(name="Default-Linebreak-Before", fontName="Helvetica", fontSize=10, leading=12, spaceBefore=12)
    )
    styles.add(
        ParagraphStyle(name="Default-Double-Linebreak", fontName="Helvetica", fontSize=10, leading=12, spaceAfter=24)
    )
    styles.add(
        ParagraphStyle(
            name="Default-Double-Linebreak-With-Before",
            fontName="Helvetica",
            fontSize=10,
            leading=12,
            spaceAfter=24,
            spaceBefore=24,
        )
    )
    styles.add(ParagraphStyle(name="Normal-Right", alignment=TA_RIGHT))
    styles.add(ParagraphStyle(name="Annotation", fontName="Helvetica", fontSize=8, leading=10))
    styles.add(
        ParagraphStyle(
            name="Annotation-Left", fontName="Helvetica", fontSize=8, leading=10, leftIndent=20, firstLineIndent=20
        )
    )
    styles.add(
        ParagraphStyle(name="Annotation-Right", fontName="Helvetica", fontSize=8, leading=10, alignment=TA_RIGHT)
    )
    return styles


# TODO to be removed
# def get_taxes(invoice, party, counterparty):
#     vat = 0
#     social_charges = 0
#     if invoice.favour == 'COUNTERPARTY':
#         vat = counterparty.accounting_information.vat if counterparty.accounting_information.vat else 0
#         social_charges = counterparty.accounting_information.social_charges if counterparty.accounting_information.social_charges else 0
#     else:
#         vat = party.vat if party is not None and party.vat else 0
#         social_charges = party.accounting_information.social_charges if party is not None and party.accounting_information.social_charges else 0
#     if invoice.vat or invoice.vat == 0:
#         vat = invoice.vat
#     if invoice.social_charges or invoice.social_charges == 0:
#         social_charges = invoice.social_charges
#     return vat, social_charges

# def banking_block(entry, invoice, styles, banking=None):
#     elements = list()
#     if banking is None:
#         bankings = entry.banking.all() if entry else None
#         if bankings:
#             if bankings.filter(currency=invoice.base_currency).exists():
#                 banking = bankings.filter(currency=invoice.base_currency).first()
#             elif bankings.filter(primary=True).exists():
#                 banking = bankings.filter(primary=True).first()
#         else:
#             return elements
#     elements.append(Paragraph('{}'.format(banking.institute), styles['Default-Linebreak-Before']))
#     if banking.institute_additional:
#         elements.append(Paragraph('{}'.format(banking.institute_additional), styles['Default']))
#     elements.append(Paragraph('IBAN: {}'.format(banking.iban), styles['Default']))
#     if banking.swift_bic:
#         elements.append(Paragraph('SWIFT: {}'.format(banking.swift_bic), styles['Default']))
#     return elements


def address_block(entry, invoice, styles, right=False, address=None, tax_id=None):
    elements = list()
    if address is None:
        addresses = entry.addresses.all() if entry is not None else None
        if addresses:
            if addresses.filter(primary=True).exists():
                address = addresses.filter(primary=True).first()
            elif addresses.count() > 0:
                address = addresses.first()
        else:
            return elements

    if right:
        default_style = styles["Default-Right"]
        default_linebreak_style = styles["Default-Right-Linebreak"]
    else:
        default_style = styles["Default"]
        default_linebreak_style = styles["Default-Linebreak"]

    casted_entry = entry.get_casted_entry()
    if entry.is_company:
        elements.append(Paragraph("<b>{0.name}</b>".format(casted_entry), default_style))
    else:
        elements.append(Paragraph("<b>{0.first_name} {0.last_name}</b>".format(casted_entry), default_style))

    elements.append(Paragraph(address.street, default_style))
    if address.street_additional is not None and address.street_additional != "":
        elements.append(Paragraph(address.street_additional, default_style))
    elements.append(Paragraph("{} {}".format(address.zip, address.geography_city), default_style))
    elements.append(Paragraph(address.geography_city.parent.parent.name, default_linebreak_style))

    if tax_id and entry.entry_accounting_information.tax_id:
        elements.append(Paragraph(f"VAT: {entry.entry_accounting_information.tax_id}", default_linebreak_style))

    return elements


def city_and_date_block(invoice, styles, entry=None, city=None, date=None):
    elements = list()
    if date is None:
        date = invoice.invoice_date
    if city is None:
        primary_address = entry.primary_address_contact() if entry else None
        if primary_address:
            city = primary_address.geography_city
    if city:
        elements.append(Paragraph("{}, {:%d.%m.%Y}".format(city, date), styles["Default-Right-Linebreak"]))
    else:
        elements.append(Paragraph("{:%d.%m.%Y}".format(date), styles["Default-Right-Linebreak"]))
    return elements


def logo_block(image):
    # print(image)
    reportlab_image = utils.ImageReader(image)
    width, height = reportlab_image.getSize()
    ratio = float(width) / float(height)
    image = Image(image, width=30 * ratio, height=30)
    return image, ratio


def signature_block(signees, styles):
    elements = list()
    if signees:
        elements.append(Paragraph("Best Regards,", styles["Default-Double-Linebreak-With-Before"]))
        for signee in signees:
            if signee is not None and signee.signature.name is not None and signee.signature.name != "":
                reportlab_image = utils.ImageReader(signee.signature)
                width, height = reportlab_image.getSize()
                ratio = float(width) / float(height)

                image = Image(signee.signature, width=50 * ratio, height=50)
                image.hAlign = "LEFT"
                elements.append(image)
                elements.append(Paragraph("{}".format(signee.computed_str), styles["Default"]))
    return elements


def add_text_block_with_context(text, invoice, styles):
    elements = list()
    if text:
        elements.append(Paragraph("", styles["Default-Linebreak"]))
        elements.append(Paragraph(Template(text).render(Context(invoice.get_context())), styles["Default-Linebreak"]))
    return elements


def add_transfer_block(invoice, receiver, styles, total_amount_gross):
    elements = list()

    if invoice.is_counterparty_invoice:
        text = f"The amount of <b>{invoice.invoice_currency.key} {format_invoice_number(total_amount_gross)}</b> will be transfered to the following bank account:"
    else:
        text = f"Please transfer <b>{invoice.invoice_currency.key} {format_invoice_number(total_amount_gross)}</b> to the following bank account:"

    elements.append(Paragraph(text, styles["Default-Linebreak-Before"]))

    if bank_account := receiver.get_banking_contact(invoice.invoice_currency):
        elements.append(Paragraph(bank_account.institute, styles["Default"]))

        if bank_account.institute_additional:
            elements.append(Paragraph(bank_account.institute_additional, styles["Default"]))

        if bank_account.iban and bank_account.iban != "":
            elements.append(Paragraph(f"IBAN: {bank_account.iban}", styles["Default"]))
        if bank_account.swift_bic:
            elements.append(Paragraph(f"SWIFT: {bank_account.swift_bic}", styles["Default"]))
        if bank_account.additional_information and bank_account.additional_information != "":
            elements.append(Paragraph(bank_account.additional_information, styles["Default"]))

    return elements


def get_gross_and_net_value(invoice):
    multiplier = 1 if not invoice.is_counterparty_invoice else -1

    total_gross = 0
    total_net = 0
    for booking_entry in invoice.booking_entries.all():
        if booking_entry.currency != invoice.invoice_currency:
            conversion_rate = booking_entry.currency.convert(booking_entry.booking_date, invoice.invoice_currency)
            total_net += multiplier * booking_entry.net_value * conversion_rate
            total_gross += multiplier * booking_entry.gross_value * conversion_rate
        else:
            total_net += multiplier * booking_entry.net_value
            total_gross += multiplier * booking_entry.gross_value

    return total_gross, total_net


def _build_booking_entry_columns(booking_entry, invoice, multiplier, styles, include_quantity=False):
    """Build columns for a single booking entry: title_col, quantity_col (optional), amount_col."""
    title_col = [Paragraph(str(booking_entry.title), styles["Annotation-Left"])]

    quantity_col = []
    if include_quantity:
        for k, v in booking_entry.parameters.get("annotation", []):
            quantity_col.append(Paragraph(f"{k}: {v}", styles["Annotation"]))

    amount_col = []
    if booking_entry.currency != invoice.invoice_currency:
        conversion_rate = booking_entry.currency.convert(booking_entry.booking_date, invoice.invoice_currency)
        title_col.append(
            Paragraph(
                f"<i>1.00 {booking_entry.currency.key} = {conversion_rate:.4f} {invoice.invoice_currency.key}</i>",
                styles["Annotation-Left"],
            )
        )
        amount_col.append(
            Paragraph(
                f"<i>{format_invoice_number(multiplier * conversion_rate * booking_entry.gross_value)} {invoice.invoice_currency.key}</i>",
                styles["Default-Right"],
            )
        )
        amount_col.append(
            Paragraph(
                f"{format_invoice_number(multiplier * booking_entry.gross_value)} {booking_entry.currency.key}",
                styles["Annotation-Right"],
            )
        )
    else:
        amount_col.append(
            Paragraph(
                f"{format_invoice_number(multiplier * booking_entry.gross_value)} {booking_entry.currency.key}",
                styles["Default-Right"],
            )
        )

    title_col.append(
        Paragraph(
            f"<i>{booking_entry.period.lower:%m.%d.%Y} - {booking_entry.period.upper:%m.%d.%Y}</i>",
            styles["Annotation-Left"],
        )
    )
    return title_col, quantity_col, amount_col


def _build_table_data(invoice, styles, vat, multiplier, include_quantity=False):
    """Build table_data for a given VAT group."""
    table_data = [
        [Paragraph("", styles["Default"])] * (3 if include_quantity else 2),
        [
            *(
                [Paragraph("Title", styles["Default-Table-Header"])]
                + ([Paragraph("End of Period", styles["Default-Table-Header"])] if include_quantity else [])
                + [Paragraph("Amount Due", styles["Default-Table-Header-Right"])]
            )
        ],
    ]
    booking_entries = invoice.booking_entries.filter(vat=vat)
    for row in booking_entries.values("account").distinct("account"):
        account = Account.objects.get(id=row["account"]) if row["account"] else None
        if account:
            table_data.append(
                [Paragraph(account.title, styles["Default"])]
                + [Paragraph("", styles["Default"])] * (2 if include_quantity else 1)
            )
        for booking_entry in booking_entries.filter(account=account).order_by("title"):
            title_col, quantity_col, amount_col = _build_booking_entry_columns(
                booking_entry, invoice, multiplier, styles, include_quantity
            )
            row = [title_col]
            if include_quantity:
                row.append(quantity_col)
            row.append(amount_col)
            table_data.append(row)

    total_gross, total_net = get_gross_and_net_value(invoice)
    table_data.append([Paragraph("", styles["Default"])] * (3 if include_quantity else 2))
    total_row = [
        [
            Paragraph("<strong>Total</strong>", styles["Default"]),
            Paragraph(f"<i>Vat: {vat * 100:.2f}%</i>", styles["Annotation"]),
            Paragraph(f"Total (excl. {vat * 100:.2f}% VAT)", styles["Annotation"]),
        ]
    ]
    if include_quantity:
        total_row.append([Paragraph("", styles["Default"])])
    total_row.append(
        [
            Paragraph(
                f"<strong>{format_invoice_number(total_gross)} {invoice.invoice_currency.key}</strong>",
                styles["Default-Right"],
            ),
            Paragraph(
                f"<i>{format_invoice_number(total_gross - total_net)} {invoice.invoice_currency.key}</i>",
                styles["Annotation-Right"],
            ),
            Paragraph(
                f"{format_invoice_number(total_net)} {invoice.invoice_currency.key}", styles["Annotation-Right"]
            ),
        ]
    )
    table_data.append(total_row)

    return table_data


def _create_booking_table(table_data, include_quantity=False):
    """Create and style a Table from data."""
    default_padding = 2 * cm
    extra_data_column_width = 3 * cm
    amount_column_width = 4.5 * cm
    if include_quantity:
        col_widths = [
            A4[0] - default_padding - extra_data_column_width - amount_column_width,
            extra_data_column_width,
            amount_column_width,
        ]
    else:
        col_widths = [A4[0] - default_padding - amount_column_width, amount_column_width]

    table = Table(table_data, colWidths=col_widths)
    table.setStyle(
        TableStyle(
            [
                ("BACKGROUND", (0, 0), (-1, 0), colors.HexColor("#f37200")),
                ("BACKGROUND", (0, 1), (-1, 1), colors.HexColor("#f38a32")),
                ("VALIGN", (0, 0), (-1, -1), "TOP"),
            ]
        )
    )
    return table


def add_booking_entries(invoice, styles, inverse=False):
    """
    Generate tables for invoice booking entries, optionally with quantity column.

    Args:
        invoice: The invoice object.
        styles: ReportLab styles dict.
        inverse: If True, use -1 multiplier (default: False).
    """
    multiplier = 1 if not invoice.is_counterparty_invoice else -1
    booking_entries = invoice.booking_entries.all()
    include_quantity = booking_entries.filter(quantity__isnull=False).exists()
    tables = []
    for vat in booking_entries.distinct("vat").values_list("vat", flat=True):
        table_data = _build_table_data(invoice, styles, vat, multiplier, include_quantity)
        table = _create_booking_table(table_data, include_quantity)
        tables.append(table)

    return tables



# Generated by Django 5.2.9 on 2026-02-12 08:49
from collections import defaultdict
from contextlib import suppress

import django.contrib.postgres.constraints
import django.contrib.postgres.fields.ranges
from django.db import migrations, models
from psycopg.types.range import DateRange
from slugify import slugify
from datetime import datetime, timedelta
import re
from django.apps import apps as django_apps
from tqdm import tqdm

from wbcore.utils.date import find_closest_to_quarter
from wbcrm.models import Product as CRMProduct

pattern = re.compile(r".*?(.*\b[A-Z]{2}[A-Z0-9]{10}\b)\s+([A-Z][a-zA-Z ]+)")
isin_regex = "([A-Z]{2}[A-Z0-9]{9}[0-9]{1})"

def migrate_booking_entry(apps, schema_editor):
    BookingEntry = apps.get_model('wbaccounting', 'BookingEntry')
    BookingEntryType = apps.get_model('wbaccounting', 'BookingEntryType')
    Invoice = apps.get_model('wbaccounting', 'Invoice')
    Invoice.objects.filter(status="PAID", payment_date__isnull=True).update(payment_date=models.F("invoice_date"))

    def get_or_create_product(product_name):
        return CRMProduct.objects.get_or_create_internal_product(product_name)[0]

    product_map = {}
    if django_apps.is_installed("wbportfolio"):
        from wbportfolio.models import Product
        def get_or_create_product(product_name: str)-> str:
            if isins := re.findall(isin_regex, product_name):
                if isins[0] in product_map:
                    return product_map[isins[0]]
                with suppress(Product.DoesNotExist):
                    product = Product.objects.get(isin=isins[0])
                    if product.accounting_product:
                        product_map[isins[0]] = product.accounting_product
                        return product.accounting_product
            return get_or_create_product(product_name)


    def _parse_key(title: str):
        m = pattern.search(title)
        if m and (product_name := m.group(1)) and (booking_type := m.group(2)):
            return get_or_create_product(product_name), BookingEntryType.objects.get_or_create(key=slugify(booking_type), defaults={"title": booking_type})[0]
        else:
            return None, None

    objs = []

    # Loop over booking and extract the proper period from its parameters (prefer) or its booking date (otherwise)
    qs = BookingEntry.objects.all()
    for booking_entry in tqdm(qs, total=qs.count()):
        product, booking_type = _parse_key(booking_entry.title)
        if product and booking_type:
            booking_entry.product_id = product.id
            booking_entry.type_id = booking_type.id
            booking_entry.key = f'product-{product.id}-type-{booking_type.id}'
        else:
            booking_entry.key = slugify(booking_entry.title)

        lower = upper = None
        # if not parameters but booking is linked to an invoice who has entries with parameters, we choose one of this as parameters, otherwise we default to booking date
        if not booking_entry.parameters:
            if d := BookingEntry.objects.filter(invoice=booking_entry.invoice, parameters__from_date__isnull=False, parameters__to_date__isnull=False).first():
                lower, upper = datetime.strptime(d.parameters["from_date"], "%d.%m.%Y"), datetime.strptime(d.parameters["to_date"], "%d.%m.%Y")
            else:
                lower, upper = booking_entry.booking_date, booking_entry.booking_date
        else:
            lower, upper = datetime.strptime(booking_entry.parameters["from_date"], "%d.%m.%Y"), datetime.strptime(booking_entry.parameters["to_date"], "%d.%m.%Y")
        if upper < lower:
            lower, upper = upper, lower
        booking_entry.period = DateRange(lower, upper + timedelta(days=1)) # we have to add +1 days because the default bound for date range field is [) and postgres does not allow changing that.
        if not booking_entry.payment_date and booking_entry.invoice:
            booking_entry.payment_date = booking_entry.invoice.payment_date
        booking_entry.parameters = {}
        objs.append(booking_entry)
    print("done loading booking entries...")
    BookingEntry.objects.bulk_update(objs, ["period", "payment_date", "product_id", "type_id", "parameters", "key"], batch_size=1000)
    print("done saving booking entries...")


def clean_overlapping_period(apps, schema_editor):
    print("cleaning up overlapping periods...")
    BookingEntry = apps.get_model('wbaccounting', 'BookingEntry')
    keys = defaultdict(set)
    entries = BookingEntry.objects.filter(account__isnull=True, invoice__isnull=True)

    # we delete not invoiced booking entry longer than half a year
    entries.annotate(duration=models.F("period__endswith") - models.F("period__startswith")).filter(duration__gt=timedelta(days=180)).delete()
    # get all duplicates that shares the same overlapping periods. Doing a raw SQL with this is less straighforward so we resort to python
    for e in tqdm(entries, total=entries.count()):
        qs = BookingEntry.objects.filter(counterparty=e.counterparty, key=e.key, period__overlap=e.period, invoice__isnull=True, account__isnull=True).exclude(id=e.id)
        if qs.exists():
            keys[(e.key, e.counterparty.id)].add(e.period)
            keys[(e.key, e.counterparty.id)].update(set(qs.values_list("period", flat=True)))
    # for each duplicated keys and its associated overlapping period, we keep the period closest to a quarter and remove the reste. This is safe as this booking do not belong to any invoice yet
    for k, periods in keys.items():
        quarter_period = find_closest_to_quarter(periods)
        entries.filter(key=k[0], counterparty_id=k[1], period__overlap=(quarter_period.lower, quarter_period.upper)).exclude(period=(quarter_period.lower, quarter_period.upper)).delete()


SQL1 = """
WITH ranked AS (
    SELECT
        id,
        ROW_NUMBER() OVER (
            PARTITION BY period, counterparty_id, key
            ORDER BY id DESC
        ) AS rn
    FROM wbaccounting_bookingentry
    WHERE account_id IS NULL AND invoice_id IS NULL
)
DELETE FROM wbaccounting_bookingentry b
USING ranked r
WHERE b.id = r.id
  AND r.rn > 1;
"""

SQL2 = """
WITH ranked AS (
    SELECT
        id,
        ROW_NUMBER() OVER (
            PARTITION BY key, period, counterparty_id, key, invoice_id
            ORDER BY id DESC
        ) AS rn
    FROM wbaccounting_bookingentry
    WHERE invoice_id IS NOT NULL AND account_id IS NULL
)
DELETE FROM wbaccounting_bookingentry b
USING ranked r
WHERE b.id = r.id
  AND r.rn > 1;
"""


BASE_DEPENDENCIES = [
        ('currency', '0002_alter_currencyfxrates_unique_together_and_more'),
        ('directory', '0015_alter_emailcontact_address_and_more'),
        ('wbaccounting', '0013_vat_alter_entryaccountinginformation_vat_vatrate'),
        ('wbcrm', '0020_alter_product_unique_together_product_unique_product'),
]
class Migration(migrations.Migration):

    dependencies = BASE_DEPENDENCIES + [("wbportfolio", "0099_product_accounting_product")] if django_apps.is_installed("wbportfolio") else BASE_DEPENDENCIES

    operations = [
        migrations.CreateModel(
            name='BookingEntryType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(db_index=True, max_length=255, unique=True, verbose_name='Key')),
                ('title', models.CharField(max_length=255, verbose_name='Title')),
            ],
            options={'verbose_name': 'Booking Type', 'verbose_name_plural': 'Bookings Type'}
        ),
        migrations.AddField(
            model_name='bookingentry',
            name='key',
            field=models.CharField(default=None, blank=True, null=True, max_length=255, verbose_name='Key'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='bookingentry',
            name='account',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='booking_entries', to='wbcrm.account', verbose_name='Account'),
        ),
        migrations.AddField(
            model_name='bookingentry',
            name='period',
            field=django.contrib.postgres.fields.ranges.DateRangeField(blank=True, null=True, verbose_name='Period'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='bookingentry',
            name='product',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='booking_entries', to='wbcrm.product', verbose_name='Product'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='bookingentry',
            name='quantity',
            field=models.IntegerField(blank=True, null=True, verbose_name='Quantity'),
        ),
        migrations.AddField(
            model_name='bookingentry',
            name='type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE,
                                    related_name='booking_entries', to='wbaccounting.bookingentrytype',
                                    verbose_name='Product'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='bookingentry',
            name='initial_booking_entry',
            field=models.ForeignKey(blank=True,
                                    help_text='If set, contains the booking entry already invoiced wich this entry values should override.',
                                    null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    related_name='related_entries', to='wbaccounting.bookingentry',
                                    verbose_name='Initial Booking Entry'),
        ),
        migrations.RunPython(migrate_booking_entry),
        migrations.RunSQL(SQL1),
        migrations.RunSQL(SQL2),
        migrations.RunPython(clean_overlapping_period)
    ]



from datetime import date
from decimal import Decimal
from unittest.mock import patch

import pytest
from psycopg.types.range import DateRange
from pytest_factoryboy import LazyFixture
from wbcore.contrib.authentication.models import User

from wbaccounting.models import BookingEntry, EntryAccountingInformation, Invoice


@pytest.mark.django_db
class TestBookingEntry:
    def test_str(self, booking_entry: BookingEntry):
        assert str(booking_entry) == f"{booking_entry.title} ({booking_entry.period})"

    @pytest.mark.parametrize(
        "method,return_value",
        [
            ("get_endpoint_basename", "wbaccounting:bookingentry"),
            ("get_representation_value_key", "id"),
            ("get_representation_label_key", "{{title}}"),
            ("get_representation_endpoint", "wbaccounting:bookingentryrepresentation-list"),
        ],
    )
    def test_wbmodel_methods(self, method: str, return_value: str):
        assert getattr(BookingEntry, method)() == return_value

    @pytest.mark.parametrize("user__is_superuser", [False])
    def test_filter_for_user_no_superuser(self, booking_entry: BookingEntry, user: User):
        booking_entries = BookingEntry.objects.filter_for_user(user)  # type: ignore
        assert booking_entry not in booking_entries

    @pytest.mark.parametrize("user__is_superuser", [True])
    def test_filter_for_user_superuser(self, booking_entry: BookingEntry, user: User):
        booking_entries = BookingEntry.objects.filter_for_user(user)  # type: ignore
        assert booking_entry in booking_entries

    def test_filter_unassigned(self, invoice, booking_entry_factory: BookingEntry):
        b1 = booking_entry_factory.create(invoice=None, payment_date=date(2025, 1, 1))  # noqa
        b2 = booking_entry_factory.create(invoice=invoice, payment_date=None)  # noqa
        b3 = booking_entry_factory.create(invoice=None, payment_date=None)

        assert set(BookingEntry.objects.filter_unassigned()) == {b3}

    @patch.object(Invoice, "get_fx_rate")
    def test_assign_invoice(self, mock_get_fx_rate, invoice, booking_entry_factory):
        mock_get_fx_rate.return_value = Decimal("1.5")
        b = booking_entry_factory.create(invoice=invoice, payment_date=None)
        assert b.invoice_fx_rate == Decimal("1.5")
        assert b.invoice == invoice
        assert b.payment_date is None

        # ensure that reassignin an invoice with a payment date set the booking entry payment date accordingly
        invoice.payment_date = date.today()
        b.assign_invoice(invoice)

        assert b.payment_date == invoice.payment_date

    @pytest.mark.parametrize("user__is_superuser", [False])
    @pytest.mark.parametrize("user__user_permissions", [(["wbaccounting.view_bookingentry"])])
    @pytest.mark.parametrize("booking_entry__counterparty", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__entry", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__counterparty_is_private", [False])
    def test_filter_for_user_public_counterparty(
        self, booking_entry: BookingEntry, user: User, entry_accounting_information: EntryAccountingInformation
    ):
        booking_entries = BookingEntry.objects.filter_for_user(user)  # type: ignore
        assert booking_entry in booking_entries

    @pytest.mark.parametrize("user__is_superuser", [False])
    @pytest.mark.parametrize("user__user_permissions", [(["wbaccounting.view_bookingentry"])])
    @pytest.mark.parametrize("booking_entry__counterparty", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__entry", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__counterparty_is_private", [True])
    def test_filter_for_user_private_counterparty(
        self, booking_entry: BookingEntry, user: User, entry_accounting_information: EntryAccountingInformation
    ):
        booking_entries = BookingEntry.objects.filter_for_user(user)  # type: ignore
        assert booking_entry not in booking_entries

    @pytest.mark.parametrize("user__is_superuser", [False])
    @pytest.mark.parametrize("user__user_permissions", [(["wbaccounting.view_bookingentry"])])
    @pytest.mark.parametrize("booking_entry__counterparty", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__entry", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__counterparty_is_private", [True])
    def test_filter_for_user_private_counterparty_with_exempt(
        self, booking_entry: BookingEntry, user: User, entry_accounting_information: EntryAccountingInformation
    ):
        entry_accounting_information.exempt_users.add(user)
        booking_entries = BookingEntry.objects.filter_for_user(user)  # type: ignore
        assert booking_entry in booking_entries

    @pytest.mark.parametrize("booking_entry__net_value", [None])
    @pytest.mark.parametrize("booking_entry__gross_value", [Decimal(99)])
    @pytest.mark.parametrize("booking_entry__vat", [Decimal(0.1)])
    def test_gross_without_net(self, booking_entry: BookingEntry):
        assert pytest.approx(booking_entry.net_value) == Decimal(90)

    def test_invoice_value_same_currency(self, booking_entry: BookingEntry):
        booking_entry.refresh_from_db()
        assert booking_entry.net_value == booking_entry.invoice_net_value
        assert booking_entry.gross_value == booking_entry.invoice_gross_value

    def test_invoice_value_different_currency(
        self, booking_entry: BookingEntry, invoice: Invoice, currency_factory, mocker
    ):
        # Change the invoice currency
        invoice.invoice_currency = currency_factory.create()
        invoice.save()

        # Patch the currency convert method to no rely on the database
        mocker.patch("wbcore.contrib.currency.models.Currency.convert", return_value=Decimal(1.1))
        # Run the save method to trigger fx computation
        booking_entry.save()
        assert booking_entry.invoice_net_value == pytest.approx(
            booking_entry.net_value * booking_entry.invoice_fx_rate
        )  # type: ignore

    # This old test that checked if save was called if booking entry was save, this was inefficent an
    # def test_invoice_save_on_change(self, booking_entry: BookingEntry, mocker):
    #     mocked_save = mocker.patch("wbaccounting.models.invoice.Invoice.save", autospec=True)
    #     booking_entry.save()
    #     mocked_save.assert_called_once()
    #
    # def test_invoice_delete_on_change(self, booking_entry: BookingEntry, mocker):
    #     mocked_save = mocker.patch("wbaccounting.models.invoice.Invoice.save", autospec=True)
    #     booking_entry.delete()
    #     mocked_save.assert_called_once()


@pytest.mark.django_db
class TestBookingEntryManager:
    def test_bulk_create_filters_out_tiny_gross_values(
        self, currency, entry, account, product_factory, booking_entry_type, booking_entry_factory
    ):
        ok_entry = booking_entry_factory.build(
            currency=currency,
            counterparty=entry,
            product=product_factory.create(),
            type=booking_entry_type,
            account=account,
            invoice=None,
            gross_value=BookingEntry.MINIMUM_VALID_GROSS_VALUE + Decimal("0.01"),
        )
        tiny_entry = booking_entry_factory.build(
            currency=currency,
            counterparty=entry,
            product=product_factory.create(),
            type=booking_entry_type,
            account=account,
            invoice=None,
            gross_value=BookingEntry.MINIMUM_VALID_GROSS_VALUE,
        )

        BookingEntry.objects.bulk_create(entry, [ok_entry, tiny_entry])

        assert BookingEntry.objects.count() == 1
        assert BookingEntry.objects.first().gross_value == ok_entry.gross_value

    def test_bulk_create_creates_basic(
        self, currency, entry, account, product, booking_entry_type, booking_entry_factory
    ):
        draft = booking_entry_factory.build(
            currency=currency,
            period=DateRange(date(2025, 1, 1), date(2025, 4, 1)),
            counterparty=entry,
            account=account,
            product=product,
            type=booking_entry_type,
            invoice=None,
            gross_value=Decimal("100"),
        )

        created = BookingEntry.objects.bulk_create(entry, [draft])

        assert len(created) == 1
        db_entry = BookingEntry.objects.get(pk=created[0].pk)
        assert db_entry.gross_value == Decimal("100")

        # assert that reimporting will update existing and delete the older one
        correction = booking_entry_factory.build(
            currency=currency,
            counterparty=entry,
            product=draft.product,
            type=draft.type,
            account=draft.account,
            period=draft.period,
            gross_value=Decimal("200"),
            invoice=None,
        )

        BookingEntry.objects.bulk_create(entry, [correction])

        # Old draft must be gone, new one present
        with pytest.raises(BookingEntry.DoesNotExist):
            draft.refresh_from_db()

        assert BookingEntry.objects.get(
            counterparty=entry,
            product=draft.product,
            type=draft.type,
            account=draft.account,
            period=draft.period,
        ).gross_value == Decimal("200")

    def test_bulk_create_if_invoiced_same_account(self, entry, booking_entry_factory, invoice_factory):
        invoice = invoice_factory()
        invoiced = booking_entry_factory.create(
            counterparty=entry,
            invoice=invoice,
            gross_value=Decimal("150"),
        )
        correction = booking_entry_factory.build(
            invoice=None,
            currency=invoiced.currency,
            counterparty=entry,
            product=invoiced.product,
            type=invoiced.type,
            account=invoiced.account,
            period=invoiced.period,
            gross_value=Decimal("200"),
        )

        created = BookingEntry.objects.bulk_create(entry, [correction])

        invoiced.refresh_from_db()
        assert created[0].gross_value == Decimal("200") - Decimal("150")
        assert invoiced.gross_value == Decimal("150")  # shouldn't be touched

    def test_bulk_create_squashable_entries_different_account(
        self, entry, invoice, account_factory, booking_entry_factory, invoice_factory
    ):
        invoiced = booking_entry_factory.create(
            counterparty=entry, invoice=invoice, gross_value=Decimal("300"), account=None
        )

        entry_1 = booking_entry_factory.build(
            invoice=None,
            currency=invoiced.currency,
            counterparty=entry,
            product=invoiced.product,
            type=invoiced.type,
            account=account_factory.create(),
            period=invoiced.period,
            gross_value=Decimal("200"),
        )

        entry_2 = booking_entry_factory.build(
            invoice=None,
            currency=invoiced.currency,
            counterparty=entry,
            product=invoiced.product,
            type=invoiced.type,
            account=account_factory.create(),
            period=invoiced.period,
            gross_value=Decimal("150"),
        )

        created = BookingEntry.objects.bulk_create(entry, [entry_1, entry_2])

        # e1/e2 should have been merged and account unset during processing
        assert len(created) == 1
        merged = created[0]
        # difference: (200 + 150) - 300 = 50
        assert merged.gross_value == Decimal("50")
        assert merged.initial_booking_entry == invoiced
        assert merged.account is None

    def test_bulk_create_creates_opposite_for_obsolete_drafts(
        self, entry, booking_entry_factory, invoice, account_factory
    ):
        ok_invoiced = booking_entry_factory.create(
            invoice=invoice, counterparty=entry, gross_value=Decimal("200"), account=None
        )
        wrongly_invoiced = booking_entry_factory.create(
            period=ok_invoiced.period, invoice=invoice, counterparty=entry, gross_value=Decimal("150"), account=None
        )

        correction = booking_entry_factory.build(
            invoice=None,
            currency=ok_invoiced.currency,
            counterparty=entry,
            product=ok_invoiced.product,
            type=ok_invoiced.type,
            account=None,
            period=ok_invoiced.period,
            gross_value=Decimal("200"),
        )

        BookingEntry.objects.bulk_create(entry, [correction])

        ok_invoiced.refresh_from_db()
        wrongly_invoiced.refresh_from_db()
        assert ok_invoiced.gross_value == Decimal("200")
        assert wrongly_invoiced.gross_value == Decimal("150")

        opposite_booking = BookingEntry.objects.get(
            counterparty=entry,
            product=wrongly_invoiced.product,
            type=wrongly_invoiced.type,
            invoice__isnull=True,
            account__isnull=True,
        )
        assert opposite_booking.gross_value == Decimal("-150")



from collections import defaultdict
from contextlib import suppress
from datetime import timedelta
from decimal import Decimal
from typing import TYPE_CHECKING

from django.contrib.postgres.constraints import ExclusionConstraint
from django.contrib.postgres.fields import DateRangeField, RangeOperators
from django.db import models, transaction
from django.dispatch import receiver
from psycopg.types.range import DateRange
from slugify import slugify
from wbcore.contrib.authentication.models import User
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.directory.models import Entry
from wbcore.models import WBModel
from wbcore.signals import pre_merge

if TYPE_CHECKING:
    from wbaccounting.models import Invoice


class BookingEntryDefaultQuerySet(models.QuerySet):
    def filter_for_user(self, user: User) -> models.QuerySet:
        """
        Filters booking entries based on if the current user can see the invoice.

        Args:
            user (User): The user for whom booking entries need to be filtered.

        Returns:
            QuerySet: A filtered queryset.
        """

        # Superuser and users with the admin permission can see all booking entries
        if user.is_superuser or user.has_perm("wbaccounting.administrate_invoice"):
            return self

        # If the user doesn't have the view permission, nothing can be seen
        if not user.has_perm("wbaccounting.view_bookingentry"):
            return self.none()

        # The user can see all booking entries where the counterparty is not private
        # or where the user is part of the exempt users list
        return self.filter(
            models.Q(counterparty__entry_accounting_information__counterparty_is_private=False)
            | models.Q(counterparty__entry_accounting_information__exempt_users=user)
        )

    def filter_unassigned(self) -> models.QuerySet:
        return self.filter(invoice__isnull=True, payment_date__isnull=True)


class BookingEntryManager(models.Manager):
    def get_queryset(self) -> BookingEntryDefaultQuerySet:
        return BookingEntryDefaultQuerySet(self.model)

    def filter_for_user(self, user: User):
        return self.get_queryset().filter_for_user(user)

    def filter_unassigned(self):
        return self.get_queryset().filter_unassigned()

    @transaction.atomic
    def bulk_create(self, counterparty: Entry, objs: list["BookingEntry"], *args, **kwargs) -> list["BookingEntry"]:
        if not objs:
            return super().bulk_create(objs, *args, **kwargs)

        created_entries = []
        squashable_entries = defaultdict(list)
        default_period = objs[0].period
        existing_invoiced_booking_entries = BookingEntry.objects.filter(
            invoice__isnull=False, period=default_period, counterparty=counterparty
        )
        for booking_entry in objs:
            conditions = (
                models.Q(counterparty=counterparty)
                & models.Q(product=booking_entry.product)
                & models.Q(type=booking_entry.type)
                & (models.Q(account=booking_entry.account) | models.Q(account__isnull=True))
                & models.Q(period__overlap=booking_entry.period)
            )
            # delete existing draft booking entry to replace them with newest values, this is safe as we exclitictly ensure these are not linked to invoice or paid yet
            BookingEntry.objects.filter(conditions).filter(invoice__isnull=True).delete()
            # We have an unpaid booking; check if there is an already-paid, invoiced one.
            try:
                invoiced_booking_entry = BookingEntry.objects.filter(conditions).get(
                    invoice__isnull=False,
                )
                booking_entry.initial_booking_entry = invoiced_booking_entry
                # if the invoiced booking entry and the newest booking entry share the same account, we can safely merge them together (by taking the gross value difference)
                if invoiced_booking_entry.account == booking_entry.account:
                    if invoiced_booking_entry.gross_value == booking_entry.gross_value:
                        continue
                    else:
                        # we store the difference from the already invoiced gross value

                        booking_entry.gross_value -= invoiced_booking_entry.gross_value
                        if booking_entry.gross_value:
                            created_entries.append(booking_entry)
                # otherwise, we have to unset the account and handle these squashable booking entries in a later state, once we have processed all entries
                else:
                    booking_entry.account = None
                    squashable_entries[invoiced_booking_entry].append(booking_entry)
                # we update fields that might have changed
                if invoiced_booking_entry.parameters != booking_entry.parameters:
                    invoiced_booking_entry.parameters = booking_entry.parameters
                    invoiced_booking_entry.quantity = booking_entry.quantity
                    invoiced_booking_entry.save()
                existing_invoiced_booking_entries = existing_invoiced_booking_entries.exclude(
                    id=invoiced_booking_entry.id
                )
            except BookingEntry.DoesNotExist:
                created_entries.append(booking_entry)

        # for each squashable booking entries (entries from an invoice whose account are not already set), we have no other choice than merging all booking entry for each account and match the sum gross value agasint the already invoiced booking entry to get the difference to be paid
        for invoiced_booking_entry, duplicated_entries in squashable_entries.items():
            total_gross_value = sum(map(lambda o: o.gross_value, duplicated_entries))
            booking_entry = duplicated_entries[
                0
            ]  # this is arbitrary but we expect all other fields to be the same for all duplicated booking entries (e.g. vat and fx rate should be equivalent among similar product, account, counterparty and type)
            booking_entry.gross_value = total_gross_value - invoiced_booking_entry.gross_value
            created_entries.append(booking_entry)

        for obselete_booking_entry in existing_invoiced_booking_entries:
            # we predelete existing draft booking entry without account as we will replace it with a new value
            BookingEntry.objects.filter(
                key=obselete_booking_entry.key,
                period=obselete_booking_entry.period,
                account__isnull=True,
                invoice__isnull=True,
                counterparty=counterparty,
            ).delete()
            created_entries.append(
                self.model(
                    key=obselete_booking_entry.key,
                    period=obselete_booking_entry.period,
                    counterparty=counterparty,
                    product=obselete_booking_entry.product,
                    type=obselete_booking_entry.type,
                    booking_date=obselete_booking_entry.booking_date,
                    vat=obselete_booking_entry.vat,
                    gross_value=obselete_booking_entry.gross_value
                    * Decimal(
                        "-1"
                    ),  # we create the opposite booking entry to represent that we overpaid a position that should have been sold
                    currency=obselete_booking_entry.currency,
                    initial_booking_entry=obselete_booking_entry,
                )
            )
        created_entries = list(
            filter(lambda o: abs(o.gross_value) > BookingEntry.MINIMUM_VALID_GROSS_VALUE, created_entries)
        )
        # Delegate: bulk insert
        return super().bulk_create(created_entries, *args, **kwargs)


class BookingEntryType(WBModel):
    key = models.CharField(max_length=255, verbose_name="Key", unique=True, db_index=True)
    title = models.CharField(max_length=255, verbose_name="Title")

    class Meta:
        verbose_name = "Booking Type"
        verbose_name_plural = "Bookings Type"

    def __str__(self) -> str:
        return self.title

    def save(self, *args, **kwargs):
        if not self.key:
            self.key = slugify(self.title)
        super().save(*args, **kwargs)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbaccounting:bookingentrytype"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{title}}"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbaccounting:bookingentrytyperepresentation-list"


class BookingEntry(WBModel):
    MINIMUM_VALID_GROSS_VALUE: int = 1  # booking entry with gross value bellow 1 are excluded

    class Meta:
        verbose_name = "Booking"
        verbose_name_plural = "Bookings"

        permissions = (
            (
                "can_generate_booking_entries",
                "Can Generate Bookings",
            ),
        )
        constraints = [
            ExclusionConstraint(
                name="exclude_overlapping_booking_entry",
                condition=models.Q(invoice__isnull=True) & models.Q(account__isnull=False),
                expressions=[
                    ("period", RangeOperators.OVERLAPS),
                    ("counterparty", RangeOperators.EQUAL),
                    ("key", RangeOperators.EQUAL),
                    ("account", RangeOperators.EQUAL),
                ],
            ),
            ExclusionConstraint(
                name="exclude_overlapping_booking_entry_no_account",
                condition=models.Q(invoice__isnull=True) & models.Q(account__isnull=True),
                expressions=[
                    ("period", RangeOperators.OVERLAPS),
                    ("counterparty", RangeOperators.EQUAL),
                    ("key", RangeOperators.EQUAL),
                ],
            ),
            ExclusionConstraint(
                name="exclude_overlapping_invoiced_booking_entry",
                condition=models.Q(invoice__isnull=False) & models.Q(account__isnull=False),
                expressions=[
                    ("period", RangeOperators.OVERLAPS),
                    ("counterparty", RangeOperators.EQUAL),
                    ("key", RangeOperators.EQUAL),
                    ("account", RangeOperators.EQUAL),
                    ("invoice", RangeOperators.EQUAL),
                ],
            ),
            ExclusionConstraint(
                name="exclude_overlapping_invoiced_booking_entry_no_account",
                condition=models.Q(invoice__isnull=False) & models.Q(account__isnull=True),
                expressions=[
                    ("period", RangeOperators.OVERLAPS),
                    ("counterparty", RangeOperators.EQUAL),
                    ("key", RangeOperators.EQUAL),
                    ("invoice", RangeOperators.EQUAL),
                ],
            ),
        ]

        indexes = [
            models.Index(fields=["counterparty", "type", "product", "account", "period"]),
            models.Index(fields=["counterparty", "type", "product", "account", "period", "invoice"]),
        ]

    def __str__(self) -> str:
        return f"{self.title} ({self.period})"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Reference date defaults to booking date if not specified
        if self.booking_date:
            if self.period is None:
                self.period = DateRange(self.booking_date, self.booking_date + timedelta(days=1))
            if not self.reference_date:
                self.reference_date = self.booking_date
        if self.product and self.type:
            self.title = self.compute_str()
        self.key = self.get_key()

    objects = BookingEntryManager()
    key = models.CharField(max_length=255, verbose_name="Key")

    # The title of the booking entry which describes what was booked
    period = DateRangeField(verbose_name="Period")
    # The entry who has to pay / receive money
    counterparty = models.ForeignKey(
        "directory.Entry", related_name="booking_entries", on_delete=models.PROTECT, verbose_name="Counterparty"
    )

    # we have to keep these fields nullable because for manually created oboking entry, we usually don't have an uniformized product and type pair
    account = models.ForeignKey(
        "wbcrm.Account",
        related_name="booking_entries",
        on_delete=models.SET_NULL,
        verbose_name="Account",
        blank=True,
        null=True,
    )

    title = models.CharField(max_length=255, verbose_name="Title")
    product = models.ForeignKey(
        "wbcrm.Product",
        related_name="booking_entries",
        on_delete=models.CASCADE,
        verbose_name="Product",
        blank=True,
        null=True,
    )
    type = models.ForeignKey(
        BookingEntryType,
        related_name="booking_entries",
        on_delete=models.CASCADE,
        verbose_name="Type",
        blank=True,
        null=True,
    )

    # The date when this booking entry is booked. Most likely the current date
    booking_date = models.DateField(verbose_name="Booking Date")

    # The date when this booking entry has to be paid / received. If None, this date is not important.
    due_date = models.DateField(null=True, blank=True, verbose_name="Due Date")

    # The date when this booking entry was paid / received. If None, this booking entry is still "open"
    payment_date = models.DateField(null=True, blank=True, verbose_name="Payment Date")

    # The reference date represents a date to which accounting period a booking entry belongs
    reference_date = models.DateField(verbose_name="Reference Date", null=True, blank=True)

    # Either Gross or Net has to be specified. If both are specifying then net is preferred.
    # If only one is specified the other one is determined based on the given VAT. The default
    # VAT is 0%. If the value of net/gross is negative, then money is paid.

    vat = models.DecimalField(max_digits=4, decimal_places=4, default=Decimal(0), verbose_name="VAT")
    gross_value = models.DecimalField(
        max_digits=16, decimal_places=4, default=Decimal("0"), verbose_name="Gross Value"
    )
    net_value = models.GeneratedField(
        expression=models.F("gross_value") / (models.Value(Decimal("1")) + models.F("vat")),
        output_field=models.DecimalField(max_digits=16, decimal_places=4),
        db_persist=True,
    )
    quantity = models.IntegerField(null=True, blank=True, verbose_name="Quantity")

    invoice_fx_rate = models.DecimalField(
        max_digits=20, decimal_places=6, null=True, blank=True, verbose_name="Invoice FX Rate"
    )
    invoice_net_value = models.GeneratedField(
        expression=models.F("invoice_fx_rate")
        * (models.F("gross_value") / (models.Value(Decimal("1")) + models.F("vat"))),
        output_field=models.DecimalField(max_digits=16, decimal_places=4, null=True, blank=True),
        db_persist=True,
    )
    invoice_gross_value = models.GeneratedField(
        expression=models.F("invoice_fx_rate") * models.F("gross_value"),
        output_field=models.DecimalField(max_digits=16, decimal_places=4, null=True, blank=True),
        db_persist=True,
    )

    # The currency of the booking entry
    currency = models.ForeignKey(
        "currency.Currency", related_name="booking_entries", on_delete=models.PROTECT, verbose_name="Currency"
    )

    invoice = models.ForeignKey(
        "Invoice",
        related_name="booking_entries",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name="Invoice",
    )
    initial_booking_entry = models.ForeignKey(
        "BookingEntry",
        related_name="related_entries",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name="Initial Booking Entry",
        help_text="If set, contains the booking entry already invoiced wich this entry values should override.",
    )
    # The backlinks stores information and the destination where to find the underlying data
    # The backlinks is rendered in lists and instances as a button
    backlinks = models.JSONField(default=dict, blank=True)

    # Extra parameters for rendering on an invoice
    parameters = models.JSONField(default=dict, blank=True)

    # Generator that produced this Booking Entry
    generator = models.CharField(max_length=256, null=True, blank=True)

    def save(self, *args, **kwargs):
        if self.invoice:
            self.assign_invoice(self.invoice)
        super().save(*args, **kwargs)

    def compute_str(self) -> str:
        return f"{self.product.title} {self.type}"

    def get_key(self) -> str:
        if self.product and self.type:
            return f"product-{self.product.id}-type-{self.type.id}"
        else:
            return slugify(self.title)

    def assign_invoice(self, invoice: "Invoice"):
        # We get the fx rate (If it is the same currency, then it will be 1)
        self.invoice = invoice
        with suppress(CurrencyFXRates.DoesNotExist):
            self.invoice_fx_rate = invoice.get_fx_rate(self.booking_date, self.currency)
            if self.invoice.payment_date and not self.payment_date:
                self.payment_date = self.invoice.payment_date

    @classmethod
    def get_endpoint_basename(cls):
        return "wbaccounting:bookingentry"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{title}}"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbaccounting:bookingentryrepresentation-list"


@receiver(pre_merge, sender="directory.Entry")
def pre_merge_entry_bookinentry(sender: models.Model, merged_object, main_object, **kwargs):
    BookingEntry.objects.filter(counterparty=merged_object).update(counterparty=main_object)



from datetime import date

from django.contrib import admin
from wbcore.contrib.directory.models import Entry

from wbaccounting.models import BookingEntry, Invoice
from wbaccounting.models.booking_entry import BookingEntryType


class BookingEntryInline(admin.TabularInline):
    model = BookingEntry
    fields = ("type", "period", "booking_date", "gross_value", "vat", "net_value")
    readonly_fields = ("net_value",)
    ordering = ("vat", "type")
    extra = 0
    show_change_link = True

    def has_add_permission(self, request, obj=None):
        return False


@admin.register(BookingEntryType)
class BookingEntryTypeModelAdmin(admin.ModelAdmin):
    pass


@admin.register(BookingEntry)
class BookingEntryModelAdmin(admin.ModelAdmin):
    list_display = (
        "product",
        "type",
        "period",
        "booking_date",
        "net_value",
        "vat",
        "currency",
        "payment_date",
        "counterparty",
        "reference_date",
    )
    search_fields = ["counterparty__computed_str"]
    raw_id_fields = ("counterparty", "currency", "invoice")
    autocomplete_fields = ("counterparty", "currency", "invoice")

    fieldsets = (
        (
            "Basic Information",
            {
                "fields": ("product", "type", "account"),
                "classes": ("wide",),
            },
        ),
        (
            "Period & Dates",
            {
                "fields": ("period", "booking_date", "due_date", "payment_date", "reference_date"),
                "classes": ("wide"),
                "description": "Key dates for the booking entry.",
            },
        ),
        (
            "Amounts",
            {
                "fields": (
                    ("gross_value", "net_value", "vat"),
                    ("quantity",),
                ),
                "classes": ("wide"),
                "description": "Gross or Net value (Net preferred if both set). Negative values indicate payment.",
            },
        ),
        (
            "Invoice Details",
            {
                "fields": (
                    ("invoice_gross_value", "invoice_net_value"),
                    ("invoice_fx_rate",),
                    "invoice",
                ),
                "classes": ("collapse",),
            },
        ),
        (
            "Parties & Currency",
            {
                "fields": ("counterparty", "currency"),
                "classes": ("wide",),
            },
        ),
        (
            "Advanced",
            {
                "fields": ("backlinks", "parameters"),
                "classes": ("collapse",),
                "description": "JSON fields for backlinks and extra parameters.",
            },
        ),
    )

    # TODO: Remove or adjust.
    def create_invoice(self, request, queryset):
        counterparty = list(set(queryset.values_list("counterparty__id", flat=True)))
        if len(counterparty) == 1:
            counterparty = Entry.objects.get(id=counterparty[0])
            invoice = Invoice.objects.create(
                title=f"Rebate Invoice {counterparty.computed_str} ({queryset.earliest('from_date').from_date:%d.%m.%Y} - {queryset.latest('to_date').to_date:%d.%m.%Y})",
                invoice_date=date.today(),
                reference_date=date.today(),
                counterparty=counterparty,
                invoice_currency=counterparty.entry_accounting_information.default_currency,
                is_counterparty_invoice=True,
            )
            for booking_entry in queryset:
                booking_entry.invoice = invoice
                booking_entry.save()
            invoice.save()

        else:
            print("-------------------------------")  # noqa: T201
            print(f"Too many counterparties selected ({len(counterparty)})")  # noqa: T201
            print("-------------------------------")  # noqa: T201

    actions = [create_invoice]



from wbcore.menus import ItemPermission, MenuItem

CASHFLOW_MENUITEM = MenuItem(
    label="Cashflow",
    endpoint="wbaccounting:futurecashflow-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal, permissions=["wbaccounting.view_transaction"]
    ),
)



from datetime import date, timedelta
from decimal import Decimal
from unittest.mock import patch

import pytest
from pytest_factoryboy.fixture import LazyFixture
from wbcore.contrib.authentication.models import User

from wbaccounting.factories import EntryAccountingInformationFactory
from wbaccounting.generators import TestGenerator
from wbaccounting.models import BookingEntry, EntryAccountingInformation, Invoice


@pytest.mark.django_db
class TestInvoice:
    def test_str(self, invoice: Invoice):
        assert str(invoice) == invoice.title

    @pytest.mark.parametrize(
        "method,return_value",
        [
            ("get_endpoint_basename", "wbaccounting:invoice"),
            ("get_representation_value_key", "id"),
            ("get_representation_label_key", "{{title}} ({{invoice_date}})"),
            ("get_representation_endpoint", "wbaccounting:invoicerepresentation-list"),
        ],
    )
    def test_wbmodel_methods(self, method: str, return_value: str):
        assert getattr(Invoice, method)() == return_value

    @pytest.mark.parametrize("user__is_superuser", [False])
    def test_filter_for_user_no_superuser(self, invoice: Invoice, user: User):
        invoices = Invoice.objects.filter_for_user(user)  # type: ignore
        assert invoice not in invoices

    @pytest.mark.parametrize("user__is_superuser", [True])
    def test_filter_for_user_superuser(self, invoice: Invoice, user: User):
        invoices = Invoice.objects.filter_for_user(user)  # type: ignore
        assert invoice in invoices

    @pytest.mark.parametrize("user__is_superuser", [False])
    @pytest.mark.parametrize("user__user_permissions", [(["wbaccounting.view_invoice"])])
    @pytest.mark.parametrize("invoice__counterparty", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__entry", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__counterparty_is_private", [False])
    def test_filter_for_user_public_counterparty(
        self, invoice: Invoice, user: User, entry_accounting_information: EntryAccountingInformation
    ):
        invoices = Invoice.objects.filter_for_user(user)  # type: ignore
        assert invoice in invoices

    @pytest.mark.parametrize("user__is_superuser", [False])
    @pytest.mark.parametrize("user__user_permissions", [(["wbaccounting.view_entryaccountinginformation"])])
    @pytest.mark.parametrize("invoice__counterparty", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__entry", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__counterparty_is_private", [True])
    def test_filter_for_user_private_counterparty(
        self, user: User, invoice: Invoice, entry_accounting_information: EntryAccountingInformation
    ):
        invoices = Invoice.objects.filter_for_user(user)  # type: ignore
        assert invoice not in invoices

    @pytest.mark.parametrize("user__is_superuser", [False])
    @pytest.mark.parametrize("user__user_permissions", [(["wbaccounting.view_entryaccountinginformation"])])
    @pytest.mark.parametrize("invoice__counterparty", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__entry", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__counterparty_is_private", [True])
    def test_filter_for_user_private_counterparty_with_exempt(
        self, invoice: Invoice, user: User, entry_accounting_information: EntryAccountingInformation
    ):
        entry_accounting_information.exempt_users.add(user)
        entry_accounting_information_list = EntryAccountingInformation.objects.filter_for_user(user)  # type: ignore
        assert entry_accounting_information in entry_accounting_information_list

    @pytest.mark.parametrize("invoice__status", [Invoice.Status.DRAFT])
    @patch.object(EntryAccountingInformation, "get_booking_entry_generator")
    @patch.object(Invoice, "get_fx_rate")
    def test_refresh_booking_entry_draft(
        self, mock_get_fx_rate, mock_get_booking_entry_generator, invoice, booking_entry_factory
    ):
        mock_get_booking_entry_generator.return_value = TestGenerator
        mock_get_fx_rate.return_value = Decimal("1")

        EntryAccountingInformationFactory.create(entry=invoice.counterparty)
        already_assigned_booking_entry = booking_entry_factory.create(
            counterparty=invoice.counterparty,
            invoice=None,
            payment_date=date(2025, 1, 1),
            reference_date=invoice.reference_date,
        )
        unassigned_booking_entry = booking_entry_factory.create(
            counterparty=invoice.counterparty, invoice=None, payment_date=None, reference_date=invoice.reference_date
        )
        booking_entry_factory.create(
            counterparty=invoice.counterparty,
            invoice=None,
            payment_date=None,
            reference_date=invoice.reference_date + timedelta(days=1),
        )
        invoiced_booking_entry = booking_entry_factory.create(
            invoice=invoice,
            counterparty=invoice.counterparty,
            payment_date=None,
            reference_date=invoice.reference_date,
        )
        correcting_booking_entry = booking_entry_factory.create(
            initial_booking_entry=invoiced_booking_entry,
            invoice=None,
            counterparty=invoice.counterparty,
            payment_date=None,
            reference_date=invoice.reference_date,
        )

        expected_new_correcting_booking_entry = (
            invoiced_booking_entry.gross_value + correcting_booking_entry.gross_value
        )
        invoice.refresh_booking_entry()

        unassigned_booking_entry.refresh_from_db()
        correcting_booking_entry.refresh_from_db()
        already_assigned_booking_entry.refresh_from_db()

        # these entries should be assigned to the invoice now
        assert unassigned_booking_entry.invoice == invoice
        assert correcting_booking_entry.invoice == invoice

        # new gross value updated based on the existing invoiced booking entry and the invoiced entry deleted
        assert correcting_booking_entry.gross_value == expected_new_correcting_booking_entry
        with pytest.raises(BookingEntry.DoesNotExist):
            invoiced_booking_entry.refresh_from_db()

        # this entry should be kept not linked to the invoice as it's already paid
        assert already_assigned_booking_entry.invoice is None

        # check that a manually addded booking entry update automatically the max reference date
        booking_entry_factory.create(invoice=invoice, reference_date=invoice.reference_date + timedelta(days=1))
        initial_reference_date = invoice.reference_date
        invoice.refresh_booking_entry()
        assert invoice.reference_date == initial_reference_date + timedelta(days=1)
